"""
tests for encutils.py
"""

import http.client
from io import StringIO

try:
    import cssutils.encutils as encutils
except ImportError:
    import encutils

# helper log
log = encutils.buildlog(stream=StringIO())


class TestAutoEncoding:
    def _fakeRes(self, content):
        "build a fake HTTP response"

        class FakeRes:
            def __init__(self, content):
                self._info = http.client.HTTPMessage()
                # Adjust to testdata.
                items = content.split(':')
                if len(items) > 1:
                    # Get the type by just
                    # using the data at the end.
                    t = items[-1].strip()
                    self._info.set_type(t)

            def info(self):
                return self._info

            def read(self):
                return content

        return FakeRes(content)

    def test_getTextTypeByMediaType(self):
        "encutils._getTextTypeByMediaType"
        tests = {
            'application/xml': encutils._XML_APPLICATION_TYPE,
            'application/xml-dtd': encutils._XML_APPLICATION_TYPE,
            'application/xml-external-parsed-entity': encutils._XML_APPLICATION_TYPE,
            'application/xhtml+xml': encutils._XML_APPLICATION_TYPE,
            'text/xml': encutils._XML_TEXT_TYPE,
            'text/xml-external-parsed-entity': encutils._XML_TEXT_TYPE,
            'text/xhtml+xml': encutils._XML_TEXT_TYPE,
            'text/html': encutils._HTML_TEXT_TYPE,
            'text/css': encutils._TEXT_UTF8,
            'text/plain': encutils._TEXT_TYPE,
            'x/x': encutils._OTHER_TYPE,
            'ANYTHING': encutils._OTHER_TYPE,
        }
        for test, exp in list(tests.items()):
            assert exp == encutils._getTextTypeByMediaType(test, log=log)

    def test_getTextType(self):
        "encutils._getTextType"
        tests = {
            '\x00\x00\xfe\xff<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '\xff\xfe\x00\x00<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '\xfe\xff<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '\xff\xfe<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '\xef\xbb\xbf<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '<?xml version="1.0"': encutils._XML_APPLICATION_TYPE,
            '\x00\x00\xfe\xffanything': encutils._OTHER_TYPE,
            '\xff\xfe\x00\x00anything': encutils._OTHER_TYPE,
            '\xfe\xffanything': encutils._OTHER_TYPE,
            '\xff\xfeanything': encutils._OTHER_TYPE,
            '\xef\xbb\xbfanything': encutils._OTHER_TYPE,
            'x/x': encutils._OTHER_TYPE,
            'ANYTHING': encutils._OTHER_TYPE,
        }
        for test, exp in list(tests.items()):
            assert exp == encutils._getTextType(test, log=log)

    def test_encodingByMediaType(self):
        "encutils.encodingByMediaType"
        tests = {
            'application/xml': 'utf-8',
            'application/xml-dtd': 'utf-8',
            'application/xml-external-parsed-entity': 'utf-8',
            'application/ANYTHING+xml': 'utf-8',
            '  application/xml  ': 'utf-8',
            'text/xml': 'ascii',
            'text/xml-external-parsed-entity': 'ascii',
            'text/ANYTHING+xml': 'ascii',
            'text/html': 'iso-8859-1',
            'text/css': 'utf-8',
            'text/plain': 'iso-8859-1',
            'ANYTHING': None,
        }
        for test, exp in list(tests.items()):
            assert exp == encutils.encodingByMediaType(test, log=log)

    def test_getMetaInfo(self):
        "encutils.getMetaInfo"
        tests = {
            """<meta tp-equiv='Content-Type' content='text/html; charset=ascii'>""": (
                None,
                None,
            ),
            """<meta http-equiv='ontent-Type' content='text/html; charset=ascii'>""": (
                None,
                None,
            ),
            """<meta http-equiv='Content-Type' content='text/html'>""": (
                'text/html',
                None,
            ),
            """<meta content='text/html' http-equiv='Content-Type'>""": (
                'text/html',
                None,
            ),
            """<meta content='text/html;charset=ascii' http-equiv='Content-Type'>""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv='Content-Type' content='text/html ;charset=ascii'>""": (
                'text/html',
                'ascii',
            ),
            """<meta content='text/html;charset=iso-8859-1' http-equiv='Content-Type'>""": (
                'text/html',
                'iso-8859-1',
            ),
            """<meta http-equiv="Content-Type" content="text/html;charset = ascii">""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv="Content-Type" content="text/html;charset=ascii;x=2">""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv="Content-Type" content="text/html;x=2;charset=ascii">""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv="Content-Type" content="text/html;x=2;charset=ascii;y=2">""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv='Content-Type' content="text/html;charset=ascii">""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv='Content-Type' content='text/html;charset=ascii'  />""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv = " Content-Type" content = " text/html;charset=ascii " >""": (
                'text/html',
                'ascii',
            ),
            """<meta http-equiv = " \n Content-Type " content = "  \t text/html   ;  charset=ascii " >""": (
                'text/html',
                'ascii',
            ),
            """<meta content="text/html;charset=ascii" http-equiv="Content-Type">""": (
                'text/html',
                'ascii',
            ),
            """<meta content="text/html;charset=ascii" http-equiv="cONTENT-type">""": (
                'text/html',
                'ascii',
            ),
            """raises exception: </ >""": (None, None),
            """<meta content="text/html;charset=ascii" http-equiv="cONTENT-type">
                </ >""": (
                'text/html',
                'ascii',
            ),
            """</ >
                <meta content="text/html;charset=ascii" http-equiv="cONTENT-type">""": (
                'text/html',
                'ascii',
            ),
            # py 2.7.3 fixed HTMLParser so:  (None, None)
            """<meta content="text/html" http-equiv="cONTENT-type">
                </ >
                <meta content="text/html;charset=ascii" http-equiv="cONTENT-type">""": (
                'text/html',
                None,
            ),
        }
        for test, exp in list(tests.items()):
            assert exp == encutils.getMetaInfo(test, log=log)

    def test_detectXMLEncoding(self):
        "encutils.detectXMLEncoding"
        tests = (
            # BOM
            (('utf_32_be'), '\x00\x00\xfe\xffanything'),
            (('utf_32_le'), '\xff\xfe\x00\x00anything'),
            (('utf_16_be'), '\xfe\xffanything'),
            (('utf_16_le'), '\xff\xfeanything'),
            (('utf-8'), '\xef\xbb\xbfanything'),
            # encoding=
            (('ascii'), '<?xml version="1.0" encoding="ascii" ?>'),
            (('ascii'), "<?xml version='1.0' encoding='ascii' ?>"),
            (('iso-8859-1'), "<?xml version='1.0' encoding='iso-8859-1' ?>"),
            # default
            (('utf-8'), '<?xml version="1.0" ?>'),
            (('utf-8'), '<?xml version="1.0"?><x encoding="ascii"/>'),
        )
        for exp, test in tests:
            assert exp == encutils.detectXMLEncoding(test, log=log)

    def test_tryEncodings(self):
        "encutils.tryEncodings"
        try:
            tests = [
                ('ascii', b'abc'),
                ('windows-1252', '€'.encode('windows-1252')),
                ('ascii', b'1'),
            ]
        except ImportError:
            tests = [
                ('ascii', b'abc'),
                ('windows-1252', '€'.encode('windows-1252')),
                ('iso-8859-1', 'äöüß'.encode('iso-8859-1')),
                ('iso-8859-1', 'äöüß'.encode('windows-1252')),
                # ('utf-8', u'\u1111'.encode('utf-8'))
            ]
        for exp, test in tests:
            assert exp == encutils.tryEncodings(test).lower()

    def test_getEncodingInfo(self):
        "encutils.getEncodingInfo"
        # (expectedencoding, expectedmismatch): (httpheader, filecontent)
        tests = [
            # --- application/xhtml+xml ---
            # header default and XML default
            (
                ('utf-8', False),
                (
                    '''Content-Type: application/xhtml+xml''',
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml"/>
                    </example>''',
                ),
            ),
            # XML default
            (
                ('utf-8', False),
                (
                    None,
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml"/>
                    </example>''',
                ),
            ),
            # meta is ignored!
            (
                ('utf-8', False),
                (
                    '''Content-Type: application/xhtml+xml''',
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml;charset=iso_M"/>
                    </example>''',
                ),
            ),
            # header enc and XML default
            (
                ('iso-h', True),
                (
                    '''Content-Type: application/xhtml+xml;charset=iso-H''',
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml"/>
                    </example>''',
                ),
            ),
            # mismatch header and XML explicit, header wins
            (
                ('iso-h', True),
                (
                    '''Content-Type: application/xhtml+xml;charset=iso-H''',
                    '''<?xml version="1.0" encoding="iso-X" ?>
                    <example/>''',
                ),
            ),
            # header == XML, meta ignored!
            (
                ('iso-h', False),
                (
                    '''Content-Type: application/xhtml+xml;charset=iso-H''',
                    '''<?xml version="1.0" encoding="iso-h" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml;charset=iso_M"/>
                    </example>''',
                ),
            ),
            # XML only, meta ignored!
            (
                ('iso-x', False),
                (
                    '''Content-Type: application/xhtml+xml''',
                    '''<?xml version="1.0" encoding="iso-X" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="application/xhtml+xml;charset=iso_M"/>
                    </example>''',
                ),
            ),
            # no text or not enough text:
            (('iso-h', False), ('Content-Type: application/xml;charset=iso-h', '1')),
            (('utf-8', False), ('Content-Type: application/xml', None)),
            ((None, False), ('Content-Type: application/xml', '1')),
            # --- text/xml ---
            # default enc
            (
                ('ascii', False),
                (
                    '''Content-Type: text/xml''',
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="text/xml"/>
                    </example>''',
                ),
            ),
            # default as XML ignored and meta completely ignored
            (
                ('ascii', False),
                (
                    '''Content-Type: text/xml''',
                    '''<?xml version="1.0" encoding="iso-X" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="text/xml;charset=iso_M"/>
                    </example>''',
                ),
            ),
            (('ascii', False), ('Content-Type: text/xml', '1')),
            (('ascii', False), ('Content-Type: text/xml', None)),
            # header enc
            (
                ('iso-h', False),
                (
                    '''Content-Type: text/xml;charset=iso-H''',
                    '''<?xml version="1.0" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="text/xml"/>
                    </example>''',
                ),
            ),
            # header only, XML and meta ignored!
            (
                ('iso-h', False),
                (
                    '''Content-Type: text/xml;charset=iso-H''',
                    '''<?xml version="1.0" encoding="iso-X" ?>
                    <example/>''',
                ),
            ),
            (
                ('iso-h', False),
                (
                    '''Content-Type: text/xml;charset=iso-H''',
                    '''<?xml version="1.0"  encoding="iso-h" ?>
                    <example>
                        <meta http-equiv="Content-Type"
                            content="text/xml;charset=iso_M"/>
                    </example>''',
                ),
            ),
            # --- text/html ---
            # default enc
            (
                ('iso-8859-1', False),
                (
                    'Content-Type: text/html;',
                    '''<meta http-equiv="Content-Type"
                                        content="text/html">''',
                ),
            ),
            (('iso-8859-1', False), ('Content-Type: text/html;', None)),
            # header enc
            (
                ('iso-h', False),
                (
                    'Content-Type: text/html;charset=iso-H',
                    '''<meta http-equiv="Content-Type"
                                    content="text/html">''',
                ),
            ),
            # meta enc
            (
                ('iso-m', False),
                (
                    'Content-Type: text/html',
                    '''<meta http-equiv="Content-Type"
                                    content="text/html;charset=iso-m">''',
                ),
            ),
            # mismatch header and meta, header wins
            (
                ('iso-h', True),
                (
                    'Content-Type: text/html;charset=iso-H',
                    '''<meta http-equiv="Content-Type"
                                    content="text/html;charset=iso-m">''',
                ),
            ),
            # no header:
            (
                (None, False),
                (
                    None,
                    '''<meta http-equiv="Content-Type"
                                content="text/html;charset=iso-m">''',
                ),
            ),
            # no encoding at all
            (
                (None, False),
                (
                    None,
                    '''<meta http-equiv="Content-Type"
                                content="text/html">''',
                ),
            ),
            ((None, False), (None, '''text''')),
            # --- no header ---
            ((None, False), (None, '')),
            (('iso-8859-1', False), ('''NoContentType''', '''OnlyText''')),
            (('iso-8859-1', False), ('Content-Type: text/html;', None)),
            (('iso-8859-1', False), ('Content-Type: text/html;', '1')),
            # XML
            (('utf-8', False), (None, '''<?xml version=''')),
            (('iso-x', False), (None, '''<?xml version="1.0" encoding="iso-X"?>''')),
            # meta ignored
            (
                ('utf-8', False),
                (
                    None,
                    '''<?xml version="1.0" ?>
                                    <html><meta http-equiv="Content-Type"
                                    content="text/html;charset=iso-m"></html>''',
                ),
            ),
            (('utf-8', False), ('Content-Type: text/css;', '1')),
            (('iso-h', False), ('Content-Type: text/css;charset=iso-h', '1')),
            # only header is used by encutils
            (('utf-8', False), ('Content-Type: text/css', '@charset "ascii";')),
        ]
        for exp, test in tests:
            header, text = test
            if header:
                res = encutils.getEncodingInfo(self._fakeRes(header), text)
            else:
                res = encutils.getEncodingInfo(text=text)

            res = (res.encoding, res.mismatch)
            assert exp == res



"""Testcases for cssutils.css.cssvariablesdelaration.CSSVariablesDeclaration."""

import xml.dom

import cssutils

from . import basetest


class TestCSSVariablesDeclaration(basetest.BaseTestCase):
    def setup_method(self):
        self.r = cssutils.css.CSSVariablesDeclaration()

    def test_init(self):
        "CSSVariablesDeclaration.__init__()"
        v = cssutils.css.CSSVariablesDeclaration()
        assert '' == v.cssText
        assert 0 == v.length
        assert v.parentRule is None

        v = cssutils.css.CSSVariablesDeclaration(cssText='x: 0')
        assert 'x: 0' == v.cssText
        assert '0' == v.getVariableValue('x')

        rule = cssutils.css.CSSVariablesRule()
        v = cssutils.css.CSSVariablesDeclaration(cssText='x: 0', parentRule=rule)
        assert rule == v.parentRule

    def test__contains__(self):
        "CSSVariablesDeclaration.__contains__(name)"
        v = cssutils.css.CSSVariablesDeclaration(cssText='x: 0; y: 2')
        for test in ('x', 'y'):
            assert test in v
            assert test.upper() in v

        assert 'z' not in v

    def test_items(self):
        "CSSVariablesDeclaration[variableName]"
        v = cssutils.css.CSSVariablesDeclaration()

        value = '0'
        v['X'] = value
        assert value == v['X']
        assert value == v.getVariableValue('X')
        assert value == v['x']
        assert value == v.getVariableValue('x')

        assert '' == v['y']
        assert '' == v.getVariableValue('y')

        v['z'] = '1'
        assert 2 == v.length

        # unsorted!
        assert sorted(v) == ['x', 'z']

        del v['z']
        assert 1 == v.length
        assert 1 == v.length

        assert '0' == v.removeVariable('x')
        assert '' == v.removeVariable('z')
        assert 0 == v.length

        v.cssText = 'x:0; y:1'
        keys = []
        # unsorted!
        for i in range(0, v.length):
            keys.append(v.item(i))
        assert sorted(keys) == ['x', 'y']

    def test_keys(self):
        "CSSVariablesDeclaration.keys()"
        v = cssutils.css.CSSVariablesDeclaration(cssText='x: 0; Y: 2')
        assert ['x', 'y'] == sorted(v.keys())

    def test_cssText(self):
        "CSSVariablesDeclaration.cssText"
        # empty
        tests = {
            '': '',
            ' ': '',
            ' \t \n  ': '',
            'x: 1': None,
            'x: "a"': None,
            'x: rgb(1, 2, 3)': None,
            'x: 1px 2px 3px': None,
            'x:1': 'x: 1',
            'x:1;': 'x: 1',
            'x  :  1  ': 'x: 1',
            'x  :  1  ;  ': 'x: 1',
            'x:1;y:2': 'x: 1;\ny: 2',
            'x:1;y:2;': 'x: 1;\ny: 2',
            'x  :  1  ;  y  :  2  ': 'x: 1;\ny: 2',
            'x  :  1  ;  y  :  2  ;  ': 'x: 1;\ny: 2',
            '/*x*/': '/*x*/',
            'x555: 5': None,
            'xxx:1;yyy:2': 'xxx: 1;\nyyy: 2',
            'xxx : 1; yyy : 2': 'xxx: 1;\nyyy: 2',
            'x:1;x:2;X:2': 'x: 2',
            'same:1;SAME:2;': 'same: 2',
            '/**/x/**/:/**/1/**/;/**/y/**/:/**/2/**/': '/**/ \n /**/ \n /**/ \n x: 1 /**/;\n/**/ \n /**/ \n /**/ \n y: 2 /**/',
        }
        self.do_equal_r(tests)

        def test_cssText2(self):
            "CSSVariablesDeclaration.cssText"

            tests = {
                'top': xml.dom.SyntaxErr,
                'top:': xml.dom.SyntaxErr,
                'top : ': xml.dom.SyntaxErr,
                'top:;': xml.dom.SyntaxErr,
                'top 0': xml.dom.SyntaxErr,
                'top 0;': xml.dom.SyntaxErr,
                ':': xml.dom.SyntaxErr,
                ':0': xml.dom.SyntaxErr,
                ':0;': xml.dom.SyntaxErr,
                ':;': xml.dom.SyntaxErr,
                ': ;': xml.dom.SyntaxErr,
                '0': xml.dom.SyntaxErr,
                '0;': xml.dom.SyntaxErr,
                ';': xml.dom.SyntaxErr,
            }
            self.do_raise_r(tests)

    def test_xVariable(self):
        "CSSVariablesDeclaration.xVariable()"
        v = cssutils.css.CSSVariablesDeclaration()
        # unset
        assert '' == v.getVariableValue('x')
        # set
        v.setVariable('x', '0')
        assert '0' == v.getVariableValue('x')
        assert '0' == v.getVariableValue('X')
        assert 'x: 0' == v.cssText
        v.setVariable('X', '0')
        assert '0' == v.getVariableValue('x')
        assert '0' == v.getVariableValue('X')
        assert 'x: 0' == v.cssText
        # remove
        assert '0' == v.removeVariable('x')
        assert '' == v.removeVariable('x')
        assert '' == v.getVariableValue('x')
        assert '' == v.cssText

    def test_imports(self):
        "CSSVariables imports"

        def fetcher(url):
            url = url.replace('\\', '/')
            url = url[url.rfind('/') + 1 :]
            return (
                None,
                {
                    '3.css': '''
                    @variables {
                        over3-2-1-0: 3;
                        over3-2-1: 3;
                        over3-2: 3;
                        over3-2-0: 3;
                        over3-1: 3;
                        over3-1-0: 3;
                        over3-0: 3;
                        local3: 3;
                    }

                ''',
                    '2.css': '''
                    @variables {
                        over3-2-1-0: 2;
                        over3-2-1: 2;
                        over3-2-0: 2;
                        over3-2: 2;
                        over2-1: 2;
                        over2-1-0: 2;
                        over2-0: 2;
                        local2: 2;
                    }

                ''',
                    '1.css': '''
                    @import "3.css";
                    @import "2.css";
                    @variables {
                        over3-2-1-0: 1;
                        over3-2-1: 1;
                        over3-1: 1;
                        over3-1-0: 1;
                        over2-1: 1;
                        over2-1-0: 1;
                        over1-0: 1;
                        local1: 1;
                    }

                ''',
                }[url],
            )

        css = '''
            @import "1.css";
            @variables {
                over3-2-1-0: 0;
                over3-2-0: 0;
                over3-1-0: 0;
                over2-1-0: 0;
                over3-0: 0;
                over2-0: 0;
                over1-0: 0;
                local0: 0;
            }
            a {
                local0: var(local0);
                local1: var(local1);
                local2: var(local2);
                local3: var(local3);
                over1-0: var(over1-0);
                over2-0: var(over2-0);
                over3-0: var(over3-0);
                over2-1: var(over2-1);
                over3-1: var(over3-1);
                over3-2: var(over3-2);
                over2-1-0: var(over2-1-0);
                over3-2-0: var(over3-2-0);
                over3-2-1: var(over3-2-1);
                over3-2-1-0: var(over3-2-1-0);
            }
        '''
        p = cssutils.CSSParser(fetcher=fetcher)
        s = p.parseString(css)

        # only these in rule of this sheet
        assert s.cssRules[1].variables.length == 8
        # but all vars in s available!
        assert s.variables.length == 15
        assert [
            'local0',
            'local1',
            'local2',
            'local3',
            'over1-0',
            'over2-0',
            'over2-1',
            'over2-1-0',
            'over3-0',
            'over3-1',
            'over3-1-0',
            'over3-2',
            'over3-2-0',
            'over3-2-1',
            'over3-2-1-0',
        ] == sorted(s.variables.keys())

        # test with variables rule
        cssutils.ser.prefs.resolveVariables = False
        assert (
            s.cssText
            == b'''@import "1.css";
@variables {
    over3-2-1-0: 0;
    over3-2-0: 0;
    over3-1-0: 0;
    over2-1-0: 0;
    over3-0: 0;
    over2-0: 0;
    over1-0: 0;
    local0: 0
    }
a {
    local0: var(local0);
    local1: var(local1);
    local2: var(local2);
    local3: var(local3);
    over1-0: var(over1-0);
    over2-0: var(over2-0);
    over3-0: var(over3-0);
    over2-1: var(over2-1);
    over3-1: var(over3-1);
    over3-2: var(over3-2);
    over2-1-0: var(over2-1-0);
    over3-2-0: var(over3-2-0);
    over3-2-1: var(over3-2-1);
    over3-2-1-0: var(over3-2-1-0)
    }'''
        )

        # test with resolved vars
        cssutils.ser.prefs.resolveVariables = True
        assert (
            s.cssText
            == b'''@import "1.css";
a {
    local0: 0;
    local1: 1;
    local2: 2;
    local3: 3;
    over1-0: 0;
    over2-0: 0;
    over3-0: 0;
    over2-1: 1;
    over3-1: 1;
    over3-2: 2;
    over2-1-0: 0;
    over3-2-0: 0;
    over3-2-1: 1;
    over3-2-1-0: 0
    }'''
        )

        s = cssutils.resolveImports(s)
        assert (
            s.cssText
            == b'''/* START @import "1.css" */
/* START @import "3.css" */
/* START @import "2.css" */
a {
    local0: 0;
    local1: 1;
    local2: 2;
    local3: 3;
    over1-0: 0;
    over2-0: 0;
    over3-0: 0;
    over2-1: 1;
    over3-1: 1;
    over3-2: 2;
    over2-1-0: 0;
    over3-2-0: 0;
    over3-2-1: 1;
    over3-2-1-0: 0
    }'''
        )

    def test_parentRule(self):
        "CSSVariablesDeclaration.parentRule"
        s = cssutils.parseString('@variables { a:1}')
        r = s.cssRules[0]
        d = r.variables
        assert r == d.parentRule

        d2 = cssutils.css.CSSVariablesDeclaration('b: 2')
        r.variables = d2
        assert r == d2.parentRule

    def test_reprANDstr(self):
        "CSSVariablesDeclaration.__repr__(), .__str__()"
        s = cssutils.css.CSSVariablesDeclaration(cssText='a:1;b:2')

        assert "2" in str(s)  # length

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)



"""Testcases for cssutils.css.CSSCharsetRule"""

import xml.dom

import pytest

import cssutils.css

from . import test_cssrule


class TestCSSCharsetRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSCharsetRule()
        self.rRO = cssutils.css.CSSCharsetRule(readonly=True)
        self.r_type = cssutils.css.CSSCharsetRule.CHARSET_RULE
        self.r_typeString = 'CHARSET_RULE'

    def test_init(self):
        "CSSCharsetRule.__init__()"
        super().test_init()
        assert self.r.encoding is None
        assert '' == self.r.cssText

        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setCssText('xxx')

    def test_InvalidModificationErr(self):
        "CSSCharsetRule InvalidModificationErr"
        self._test_InvalidModificationErr('@charset')

    def test_init_encoding(self):
        "CSSCharsetRule.__init__(encoding)"
        for enc in (None, 'UTF-8', 'utf-8', 'iso-8859-1', 'ascii'):
            r = cssutils.css.CSSCharsetRule(enc)
            if enc is None:
                assert r.encoding is None
                assert '' == r.cssText
            else:
                assert enc.lower() == r.encoding
                assert '@charset "%s";' % enc.lower() == r.cssText

        for enc in (' ascii ', ' ascii', 'ascii '):
            with pytest.raises(xml.dom.SyntaxErr, match="Syntax Error"):
                cssutils.css.CSSCharsetRule(enc)

        for enc in ('unknown',):
            with pytest.raises(xml.dom.SyntaxErr, match=r"Unknown \(Python\) encoding"):
                cssutils.css.CSSCharsetRule(enc)

    def test_encoding(self):
        "CSSCharsetRule.encoding"
        for enc in ('UTF-8', 'utf-8', 'iso-8859-1', 'ascii'):
            self.r.encoding = enc
            assert enc.lower() == self.r.encoding
            assert '@charset "%s";' % enc.lower() == self.r.cssText

        for enc in (None, ' ascii ', ' ascii', 'ascii '):
            with pytest.raises(xml.dom.SyntaxErr, match="Syntax Error"):
                self.r.encoding = enc

        for enc in ('unknown',):
            with pytest.raises(xml.dom.SyntaxErr, match=r"Unknown \(Python\) encoding"):
                self.r.encoding = enc

    def test_cssText(self):
        """CSSCharsetRule.cssText

        setting cssText is ok to use @CHARSET or other but a file
        using parse MUST use ``@charset "ENCODING";``
        """
        tests = {
            '@charset "utf-8";': None,
            "@charset 'utf-8';": '@charset "utf-8";',
        }
        self.do_equal_r(tests)
        self.do_equal_p(tests)  # also parse

        tests = {
            # token is "@charset " with space!
            '@charset;"': xml.dom.InvalidModificationErr,
            '@CHARSET "UTF-8";': xml.dom.InvalidModificationErr,
            '@charset "";': xml.dom.SyntaxErr,
            '''@charset /*1*/"utf-8"/*2*/;''': xml.dom.SyntaxErr,
            '''@charset /*1*/"utf-8";''': xml.dom.SyntaxErr,
            '''@charset "utf-8"/*2*/;''': xml.dom.SyntaxErr,
            '@charset { utf-8 }': xml.dom.SyntaxErr,
            '@charset "utf-8"': xml.dom.SyntaxErr,
            '@charset a;': xml.dom.SyntaxErr,
            '@charset /**/;': xml.dom.SyntaxErr,
            # trailing content
            '@charset "utf-8";s': xml.dom.SyntaxErr,
            '@charset "utf-8";/**/': xml.dom.SyntaxErr,
            '@charset "utf-8"; ': xml.dom.SyntaxErr,
            # comments do not work in this rule!
            '@charset "utf-8"/*1*//*2*/;': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests)

    def test_repr(self):
        "CSSCharsetRule.__repr__()"
        self.r.encoding = 'utf-8'
        assert 'utf-8' in repr(self.r)

    def test_reprANDstr(self):
        "CSSCharsetRule.__repr__(), .__str__()"
        encoding = 'utf-8'

        s = cssutils.css.CSSCharsetRule(encoding=encoding)

        assert encoding in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert encoding == s2.encoding



"""CSSImportRule implements DOM Level 2 CSS CSSImportRule plus the
``name`` property from http://www.w3.org/TR/css3-cascade/#cascading."""

__all__ = ['CSSImportRule']

import os
import urllib.parse
import xml.dom

import cssutils

from . import cssrule


class CSSImportRule(cssrule.CSSRule):
    """
    Represents an @import rule within a CSS style sheet.  The @import rule
    is used to import style rules from other style sheets.

    Format::

        import
          : IMPORT_SYM S*
          [STRING|URI] S* [ medium [ COMMA S* medium]* ]? S* STRING? S* ';' S*
          ;
    """

    def __init__(
        self,
        href=None,
        mediaText=None,
        name=None,
        parentRule=None,
        parentStyleSheet=None,
        readonly=False,
    ):
        """
        If readonly allows setting of properties in constructor only

        :param href:
            location of the style sheet to be imported.
        :param mediaText:
            A list of media types for which this style sheet may be used
            as a string
        :param name:
            Additional name of imported style sheet
        """
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)
        self._atkeyword = '@import'
        self._styleSheet = None

        # string or uri used for reserialization
        self.hreftype = None

        # prepare seq
        seq = self._tempSeq()
        seq.append(None, 'href')
        # seq.append(None, 'media')
        seq.append(None, 'name')
        self._setSeq(seq)

        # 1. media
        if mediaText:
            self.media = mediaText
        else:
            # must be all for @import
            self.media = cssutils.stylesheets.MediaList(mediaText='all')
        # 2. name
        self.name = name
        # 3. href and styleSheet
        self.href = href

        self._readonly = readonly

    def __repr__(self):
        if self._usemedia:
            mediaText = self.media.mediaText
        else:
            mediaText = None
        return f"cssutils.css.{self.__class__.__name__}(href={self.href!r}, mediaText={mediaText!r}, name={self.name!r})"

    def __str__(self):
        if self._usemedia:
            mediaText = self.media.mediaText
        else:
            mediaText = None
        return f"<cssutils.css.{self.__class__.__name__} object href={self.href!r} mediaText={mediaText!r} name={self.name!r} at 0x{id(self):x}>"

    _usemedia = property(
        lambda self: self.media.mediaText not in ('', 'all'),
        doc="if self.media is used (or simply empty)",
    )

    def _getCssText(self):
        """Return serialized property cssText."""
        return cssutils.ser.do_CSSImportRule(self)

    def _setCssText(self, cssText):  # noqa: C901
        """
        :exceptions:
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
        """
        super()._setCssText(cssText)
        tokenizer = self._tokenize2(cssText)
        attoken = self._nexttoken(tokenizer, None)
        if self._type(attoken) != self._prods.IMPORT_SYM:
            self._log.error(
                'CSSImportRule: No CSSImportRule found: %s' % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            # for closures: must be a mutable
            new = {
                'keyword': self._tokenvalue(attoken),
                'href': None,
                'hreftype': None,
                'media': None,
                'name': None,
                'wellformed': True,
            }

            def __doname(seq, token):
                # called by _string or _ident
                new['name'] = self._stringtokenvalue(token)
                seq.append(new['name'], 'name')
                return ';'

            def _string(expected, seq, token, tokenizer=None):
                if 'href' == expected:
                    # href
                    new['href'] = self._stringtokenvalue(token)
                    new['hreftype'] = 'string'
                    seq.append(new['href'], 'href')
                    return 'media name ;'
                elif 'name' in expected:
                    # name
                    return __doname(seq, token)
                else:
                    new['wellformed'] = False
                    self._log.error('CSSImportRule: Unexpected string.', token)
                    return expected

            def _uri(expected, seq, token, tokenizer=None):
                # href
                if 'href' == expected:
                    uri = self._uritokenvalue(token)
                    new['hreftype'] = 'uri'
                    new['href'] = uri
                    seq.append(new['href'], 'href')
                    return 'media name ;'
                else:
                    new['wellformed'] = False
                    self._log.error('CSSImportRule: Unexpected URI.', token)
                    return expected

            def _ident(expected, seq, token, tokenizer=None):
                # medialist ending with ; which is checked upon too
                if expected.startswith('media'):
                    mediatokens = self._tokensupto2(
                        tokenizer, importmediaqueryendonly=True
                    )
                    mediatokens.insert(0, token)  # push found token

                    last = mediatokens.pop()  # retrieve ;
                    lastval, lasttyp = self._tokenvalue(last), self._type(last)
                    if lastval != ';' and lasttyp not in ('EOF', self._prods.STRING):
                        new['wellformed'] = False
                        self._log.error(
                            'CSSImportRule: No ";" found: %s' % self._valuestr(cssText),
                            token=token,
                        )

                    newMedia = cssutils.stylesheets.MediaList(parentRule=self)
                    newMedia.mediaText = mediatokens
                    if newMedia.wellformed:
                        new['media'] = newMedia
                        seq.append(newMedia, 'media')
                    else:
                        new['wellformed'] = False
                        self._log.error(
                            'CSSImportRule: Invalid MediaList: %s'
                            % self._valuestr(cssText),
                            token=token,
                        )

                    if lasttyp == self._prods.STRING:
                        # name
                        return __doname(seq, last)
                    else:
                        return 'EOF'  # ';' is token "last"
                else:
                    new['wellformed'] = False
                    self._log.error('CSSImportRule: Unexpected ident.', token)
                    return expected

            def _char(expected, seq, token, tokenizer=None):
                # final ;
                val = self._tokenvalue(token)
                if expected.endswith(';') and ';' == val:
                    return 'EOF'
                else:
                    new['wellformed'] = False
                    self._log.error('CSSImportRule: Unexpected char.', token)
                    return expected

            # import : IMPORT_SYM S* [STRING|URI]
            #            S* [ medium [ ',' S* medium]* ]? ';' S*
            #         STRING? # see http://www.w3.org/TR/css3-cascade/#cascading
            #        ;
            newseq = self._tempSeq()
            wellformed, expected = self._parse(
                expected='href',
                seq=newseq,
                tokenizer=tokenizer,
                productions={
                    'STRING': _string,
                    'URI': _uri,
                    'IDENT': _ident,
                    'CHAR': _char,
                },
                new=new,
            )

            # wellformed set by parse
            ok = wellformed and new['wellformed']

            # post conditions
            if not new['href']:
                ok = False
                self._log.error(
                    'CSSImportRule: No href found: %s' % self._valuestr(cssText)
                )

            if expected != 'EOF':
                ok = False
                self._log.error(
                    'CSSImportRule: No ";" found: %s' % self._valuestr(cssText)
                )

            # set all
            if ok:
                self._setSeq(newseq)

                self.atkeyword = new['keyword']
                self.hreftype = new['hreftype']
                self.name = new['name']

                if new['media']:
                    self.media = new['media']
                else:
                    # must be all for @import
                    self.media = cssutils.stylesheets.MediaList(mediaText='all')

                # needs new self.media
                self.href = new['href']

    cssText = property(
        fget=_getCssText,
        fset=_setCssText,
        doc="(DOM) The parsable textual representation of this rule.",
    )

    def _setHref(self, href):
        # set new href
        self._href = href
        # update seq
        for i, item in enumerate(self.seq):
            type_ = item.type
            if 'href' == type_:
                self._seq[i] = (href, type_, item.line, item.col)
                break

        importedSheet = cssutils.css.CSSStyleSheet(
            media=self.media, ownerRule=self, title=self.name
        )
        self.hrefFound = False
        # set styleSheet
        if href and self.parentStyleSheet:
            # loading errors are all catched!

            # relative href
            parentHref = self.parentStyleSheet.href
            if parentHref is None:
                # use cwd instead
                parentHref = cssutils.helper.path2url(os.getcwd()) + '/'

            fullhref = urllib.parse.urljoin(parentHref, self.href)

            # all possible exceptions are ignored
            try:
                usedEncoding, enctype, cssText = self.parentStyleSheet._resolveImport(
                    fullhref
                )

                if cssText is None:
                    # catched in next except below!
                    raise OSError('Cannot read Stylesheet.')

                # contentEncoding with parentStyleSheet.overrideEncoding,
                # HTTP or parent
                encodingOverride, encoding = None, None

                if enctype == 0:
                    encodingOverride = usedEncoding
                elif 0 < enctype < 5:
                    encoding = usedEncoding

                # inherit fetcher for @imports in styleSheet
                importedSheet._href = fullhref
                importedSheet._setFetcher(self.parentStyleSheet._fetcher)
                importedSheet._setCssTextWithEncodingOverride(
                    cssText, encodingOverride=encodingOverride, encoding=encoding
                )

            except (OSError, ValueError) as e:
                self._log.warn(
                    'CSSImportRule: While processing imported '
                    'style sheet href=%s: %r' % (self.href, e),
                    neverraise=True,
                )

            else:
                # used by resolveImports if to keep unprocessed href
                self.hrefFound = True

        self._styleSheet = importedSheet

    _href = None  # needs to be set
    href = property(
        lambda self: self._href,
        _setHref,
        doc="Location of the style sheet to be imported.",
    )

    def _setMedia(self, media):
        """
        :param media:
            a :class:`~cssutils.stylesheets.MediaList` or string
        """
        self._checkReadonly()
        if isinstance(media, str):
            self._media = cssutils.stylesheets.MediaList(
                mediaText=media, parentRule=self
            )
        else:
            media._parentRule = self
            self._media = media

        # update seq
        ihref = 0
        for i, item in enumerate(self.seq):
            if item.type == 'href':
                ihref = i
            elif item.type == 'media':
                self.seq[i] = (self._media, 'media', None, None)
                break
        else:
            # if no media until now add after href
            self.seq.insert(ihref + 1, self._media, 'media', None, None)

    media = property(
        lambda self: self._media,
        _setMedia,
        doc="(DOM) A list of media types for this rule "
        "of type :class:`~cssutils.stylesheets.MediaList`.",
    )

    def _setName(self, name=''):
        """Raises xml.dom.SyntaxErr if name is not a string."""
        if name is None or isinstance(name, str):
            # "" or '' handled as None
            if not name:
                name = None

            # save name
            self._name = name

            # update seq
            for i, item in enumerate(self.seq):
                typ = item.type
                if 'name' == typ:
                    self._seq[i] = (name, typ, item.line, item.col)
                    break

            # set title of imported sheet
            if self.styleSheet:
                self.styleSheet.title = name

        else:
            self._log.error('CSSImportRule: Not a valid name: %s' % name)

    def _getName(self):
        return self._name

    name = property(
        _getName,
        _setName,
        doc="An optional name for the imported sheet.",
    )

    styleSheet = property(
        lambda self: self._styleSheet,
        doc="(readonly) The style sheet referred to by this " "rule.",
    )

    type = property(
        lambda self: self.IMPORT_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    def _getWellformed(self):
        "Depending on if media is used at all."
        if self._usemedia:
            return bool(self.href and self.media.wellformed)
        else:
            return bool(self.href)

    wellformed = property(_getWellformed)



"""Tests for css.CSSStyleSheet"""

import xml.dom

import pytest

import cssutils.css

from . import basetest


class TestCSSStyleSheet(basetest.BaseTestCase):
    def setup_method(self):
        self.r = cssutils.css.CSSStyleSheet()  # used by basetest
        self.s = self.r  # used here
        self.rule = cssutils.css.CSSStyleRule()

    def test_init(self):
        "CSSStyleSheet.__init__()"
        assert 'text/css' == self.s.type
        assert not self.s._readonly
        assert [] == self.s.cssRules
        assert not self.s.disabled
        assert self.s.href is None
        assert self.s.media is None
        assert self.s.ownerNode is None
        assert self.s.parentStyleSheet is None
        assert '' == self.s.title

        # check that type is readonly
        with pytest.raises(AttributeError):
            self.r.__setattr__('href', 'x')
        with pytest.raises(AttributeError):
            self.r.__setattr__('parentStyleSheet', 'x')
        with pytest.raises(AttributeError):
            self.r.__setattr__('ownerNode', 'x')
        with pytest.raises(AttributeError):
            self.r.__setattr__('type', 'x')

    def test_iter(self):
        "CSSStyleSheet.__iter__()"
        s = cssutils.css.CSSStyleSheet()
        s.cssText = '''@import "x";@import "y";@namespace "u";'''
        types = [
            cssutils.css.CSSRule.IMPORT_RULE,
            cssutils.css.CSSRule.IMPORT_RULE,
            cssutils.css.CSSRule.NAMESPACE_RULE,
        ]
        for i, rule in enumerate(s):
            assert rule == s.cssRules[i]
            assert rule.type == types[i]

    def test_refs(self):
        """CSSStylesheet references"""
        s = cssutils.parseString('a {}')
        rules = s.cssRules
        assert s.cssRules[0].parentStyleSheet == s
        assert rules[0].parentStyleSheet == s

        # set cssText
        s.cssText = 'b{}'

        # from 0.9.7b1
        assert rules != s.cssRules

        # set cssRules
        s.cssRules = cssutils.parseString(
            '''
            @charset "ascii";
            /**/
            @import "x";
            @namespace "http://example.com/ns0";
            @media all {
                a { color: green; }
                }
            @font-face {
                font-family: x;
                }
            @page {
                font-family: Arial;
                }
            @x;
            b {}').cssRules'''
        ).cssRules
        # new object
        assert rules != s.cssRules
        for r in s.cssRules:
            assert r.parentStyleSheet == s

        # namespaces
        s = cssutils.parseString('@namespace "http://example.com/ns1"; a {}')
        namespaces = s.namespaces
        assert list(s.namespaces.items()) == [('', 'http://example.com/ns1')]
        s.cssText = '@namespace x "http://example.com/ns2"; x|a {}'
        # not anymore!
        assert namespaces != s.namespaces
        assert list(s.namespaces.items()) == [('x', 'http://example.com/ns2')]

        # variables
        s = cssutils.parseString('@variables { a:1}')
        vars1 = s.variables
        assert vars1['a'] == '1'

        s = cssutils.parseString('@variables { a:2}')
        vars2 = s.variables
        assert vars1 != vars2
        assert vars1['a'] == '1'
        assert vars2['a'] == '2'

    def test_cssRules(self):
        "CSSStyleSheet.cssRules"
        s = cssutils.parseString('/*1*/a {x:1}')
        assert 2 == s.cssRules.length
        del s.cssRules[0]
        assert 1 == s.cssRules.length
        s.cssRules.append('/*2*/')
        assert 2 == s.cssRules.length
        s.cssRules.extend(cssutils.parseString('/*3*/x {y:2}').cssRules)
        assert 4 == s.cssRules.length
        assert b'a {\n    x: 1\n    }\n/*2*/\n/*3*/\nx {\n    y: 2\n    }' == s.cssText

        for r in s.cssRules:
            assert r.parentStyleSheet == s

    def test_cssText(self):
        "CSSStyleSheet.cssText"
        tests = {
            '': b'',
            # @charset
            '@charset "ascii";\n@import "x";': b'@charset "ascii";\n@import "x";',
            '@charset "ascii";\n@media all {}': b'@charset "ascii";',
            '@charset "ascii";\n@x;': b'@charset "ascii";\n@x;',
            '@charset "ascii";\na {\n    x: 1\n    }': b'@charset "ascii";\na {\n    x: 1\n    }',
            # @import
            '@x;\n@import "x";': b'@x;\n@import "x";',
            '@import "x";\n@import "y";': b'@import "x";\n@import "y";',
            '@import "x";\n@media all {}': b'@import "x";',
            '@import "x";\n@x;': b'@import "x";\n@x;',
            '@import "x";\na {\n    x: 1\n    }': b'@import "x";\na {\n    x: 1\n    }',
            # @namespace
            '@x;\n@namespace a "x";': b'@x;\n@namespace a "x";',
            '@namespace a "x";\n@namespace b "y";': b'@namespace a "x";\n@namespace b "y";',
            '@import "x";\n@namespace a "x";\n@media all {}': b'@import "x";\n@namespace a "x";',
            '@namespace a "x";\n@x;': b'@namespace a "x";\n@x;',
            '@namespace a "x";\na {\n    x: 1\n    }': b'@namespace a "x";\na {\n    x: 1\n    }',
            """@namespace url("e1");
                @namespace url("e2");
                @namespace x url("x1");
                @namespace x url("x2");
                test{color: green}
                x|test {color: green}""": b"""@namespace "e2";
@namespace x "x2";
test {
    color: green
    }
x|test {
    color: green
    }""",
            #            ur'\1 { \2: \3 }': ur'''\x01 {
            #    \x02: \x03
            #    }''',
            #            ur'''
            #            \@ { \@: \@ }
            #            \1 { \2: \3 }
            #            \{{\::\;;}
            #            ''': ur'''\@ {
            #    \@: \@
            #    }
            # \1 {
            #    \2: \3
            #    }
            # \{
            #    {\:: \;
            #    }'''
        }
        self.do_equal_r(tests)

        tests = {
            '': None,
            # @charset
            '@charset "ascii";\n@import "x";': None,
            '@charset "ascii";\n@media all {}': '@charset "ascii";',
            '@charset "ascii";\n@x;': None,
            '@charset "ascii";\na {\n    x: 1\n    }': None,
            # @import
            '@x;\n@import "x";': None,
            '@import "x";\n@import "y";': None,
            '@import "x";\n@media all {}': '@import "x";',
            '@import "x";\n@x;': None,
            '@import "x";\na {\n    x: 1\n    }': None,
            # @namespace
            '@x;\n@namespace a "x";': None,
            '@namespace a "x";\n@namespace b "y";': None,
            '@import "x";\n@namespace a "x";\n@media all {}': '@import "x";\n@namespace a "x";',
            '@namespace a "x";\n@x;': None,
            '@namespace a "x";\na {\n    x: 1\n    }': None,
            """@namespace url("e1");
                @namespace url("e2");
                @namespace x url("x1");
                @namespace x url("x2");
                test{color: green}
                x|test {color: green}""": """@namespace "e2";
@namespace x "x2";
test {
    color: green
    }
x|test {
    color: green
    }""",
            #            ur'\1 { \2: \3 }': ur'''\x01 {
            #    \x02: \x03
            #    }''',
            #            ur'''
            #            \@ { \@: \@ }
            #            \1 { \2: \3 }
            #            \{{\::\;;}
            #            ''': ur'''\@ {
            #    \@: \@
            #    }
            # \1 {
            #    \2: \3
            #    }
            # \{
            #    {\:: \;
            #    }'''
        }
        self.do_equal_p(tests)

        s = cssutils.css.CSSStyleSheet()
        s.cssText = '''@charset "ascii";@import "x";@namespace a "x";
        @media all {/*1*/}@page {margin: 0}a {\n    x: 1\n    }@unknown;/*comment*/'''
        for r in s.cssRules:
            assert s == r.parentStyleSheet

    def test_cssText_HierarchyRequestErr(self):
        "CSSStyleSheet.cssText HierarchyRequestErr"
        tests = {
            # @charset: only one and always 1st
            ' @charset "utf-8";': xml.dom.HierarchyRequestErr,
            '@charset "ascii";@charset "ascii";': xml.dom.HierarchyRequestErr,
            '/*c*/@charset "ascii";': xml.dom.HierarchyRequestErr,
            '@import "x"; @charset "ascii";': xml.dom.HierarchyRequestErr,
            '@namespace a "x"; @charset "ascii";': xml.dom.HierarchyRequestErr,
            '@media all {} @charset "ascii";': xml.dom.HierarchyRequestErr,
            '@page {} @charset "ascii";': xml.dom.HierarchyRequestErr,
            'a {} @charset "ascii";': xml.dom.HierarchyRequestErr,
            # @import: before @namespace, @media, @page, sr
            '@namespace a "x"; @import "x";': xml.dom.HierarchyRequestErr,
            '@media all {} @import "x";': xml.dom.HierarchyRequestErr,
            '@page {} @import "x";': xml.dom.HierarchyRequestErr,
            'a {} @import "x";': xml.dom.HierarchyRequestErr,
            # @namespace: before @media, @page, sr
            '@media all {} @namespace a "x";': xml.dom.HierarchyRequestErr,
            '@page {} @namespace a "x";': xml.dom.HierarchyRequestErr,
            'a {} @namespace a "x";': xml.dom.HierarchyRequestErr,
        }
        self.do_raise_r(tests)
        self.do_raise_p(tests)

    def test_cssText_SyntaxErr(self):
        """CSSStyleSheet.cssText SyntaxErr

        for single {, } or ;
        """
        tests = {
            '{': xml.dom.SyntaxErr,
            '}': xml.dom.SyntaxErr,
            ';': xml.dom.SyntaxErr,
            '@charset "ascii";{': xml.dom.SyntaxErr,
            '@charset "ascii";}': xml.dom.SyntaxErr,
            '@charset "ascii";;': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests)
        self.do_raise_p(tests)

    def test_encoding(self):
        "CSSStyleSheet.encoding"
        self.s.cssText = ''
        assert 'utf-8' == self.s.encoding

        self.s.encoding = 'ascii'
        assert 'ascii' == self.s.encoding
        assert 1 == self.s.cssRules.length
        assert 'ascii' == self.s.cssRules[0].encoding

        self.s.encoding = None
        assert 'utf-8' == self.s.encoding
        assert 0 == self.s.cssRules.length

        self.s.encoding = 'UTF-8'
        assert 'utf-8' == self.s.encoding
        assert 1 == self.s.cssRules.length

        with pytest.raises(xml.dom.SyntaxErr):
            self.s._setEncoding('INVALID ENCODING')
        assert 'utf-8' == self.s.encoding
        assert 1 == self.s.cssRules.length

    def test_namespaces1(self):
        "CSSStyleSheet.namespaces.namespaces"
        # tests for namespaces internal methods
        s = cssutils.css.CSSStyleSheet()
        assert 0 == len(s.namespaces)

        css = '''@namespace "default";
@namespace ex "example";
@namespace ex2 "example";
ex2|x { top: 0 }'''
        expcss = '''@namespace "default";
@namespace ex2 "example";
ex2|x {
    top: 0
    }'''
        s.cssText = css
        assert s.cssText == expcss.encode()
        assert s.namespaces.namespaces == {'': 'default', 'ex2': 'example'}

        # __contains__
        assert '' in s.namespaces
        assert 'ex2' in s.namespaces
        assert "NOTSET" not in s.namespaces
        # __delitem__
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            s.namespaces.__delitem__('ex2')
        s.namespaces['del'] = 'del'
        del s.namespaces['del']
        with pytest.raises(xml.dom.NamespaceErr):
            s.namespaces.__getitem__('del')
        # __getitem__
        assert 'default' == s.namespaces['']
        assert 'example' == s.namespaces['ex2']
        with pytest.raises(xml.dom.NamespaceErr):
            s.namespaces.__getitem__('UNSET')
        # __iter__
        assert ['', 'ex2'] == sorted(list(s.namespaces))
        # __len__
        assert 2 == len(s.namespaces)
        # __setitem__
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            s.namespaces.__setitem__('ex2', 'NEWURI')
        s.namespaces['n1'] = 'new'
        assert s.namespaces.namespaces == {'': 'default', 'ex2': 'example', 'n1': 'new'}
        s.namespaces['n'] = 'new'  # replaces prefix!
        assert s.namespaces.namespaces == {'': 'default', 'ex2': 'example', 'n': 'new'}
        # prefixForNamespaceURI
        assert '' == s.namespaces.prefixForNamespaceURI('default')
        assert 'ex2' == s.namespaces.prefixForNamespaceURI('example')
        with pytest.raises(IndexError):
            s.namespaces.prefixForNamespaceURI('UNSET')
        # .keys
        assert set(s.namespaces.keys()) == {'', 'ex2', 'n'}
        # .get
        assert 'x' == s.namespaces.get('UNKNOWN', 'x')
        assert 'example' == s.namespaces.get('ex2', 'not used defa')

    def test_namespaces2(self):
        "CSSStyleSheet.namespaces"
        # tests using CSSStyleSheet.namespaces

        s = cssutils.css.CSSStyleSheet()
        css = '@namespace n "new";'
        # doubles will be removed
        s.insertRule(css + css)
        assert s.cssText == css.encode()
        r = cssutils.css.CSSNamespaceRule(prefix='ex2', namespaceURI='example')
        s.insertRule(r)
        r = cssutils.css.CSSNamespaceRule(namespaceURI='default')
        s.insertRule(r)

        expcss = '''@namespace n "new";
@namespace ex2 "example";
@namespace "default";'''
        assert s.cssText == expcss.encode()
        r.prefix = 'DEFAULT'
        expcss = '''@namespace n "new";
@namespace ex2 "example";
@namespace DEFAULT "default";'''
        assert s.cssText == expcss.encode()

        # CSSMediaRule
        with pytest.raises(xml.dom.NamespaceErr):
            s.add('@media all {x|a {left: 0}}')
        mcss = '@media all {\n    ex2|SEL1 {\n        left: 0\n        }\n    }'
        s.add(mcss)
        expcss += '\n' + mcss
        assert s.cssText == expcss.encode()

        # CSSStyleRule
        with pytest.raises(xml.dom.NamespaceErr):
            s.add('x|a {top: 0}')
        scss = 'n|SEL2 {\n    top: 0\n    }'
        s.add(scss)
        expcss += '\n' + scss
        assert s.cssText == expcss.encode()

        mr = s.cssRules[3]
        sr = s.cssRules[4]

        # SelectorList @media
        with pytest.raises(xml.dom.NamespaceErr):
            mr.cssRules[0]._setSelectorText('x|b')
        oldsel, newsel = mr.cssRules[0].selectorText, 'n|SEL3, a'
        mr.cssRules[0].selectorText = newsel
        expcss = expcss.replace(oldsel, newsel)
        assert s.cssText == expcss.encode()
        # SelectorList stylerule
        with pytest.raises(xml.dom.NamespaceErr):
            sr._setSelectorText('x|b')
        oldsel, newsel = sr.selectorText, 'ex2|SEL4, a'
        sr.selectorText = newsel
        expcss = expcss.replace(oldsel, newsel)
        assert s.cssText == expcss.encode()

        # Selector @media
        with pytest.raises(xml.dom.NamespaceErr):
            mr.cssRules[0].selectorList.append('x|b')
        oldsel, newsel = mr.cssRules[0].selectorText, 'ex2|SELMR'
        mr.cssRules[0].selectorList.append(newsel)
        expcss = expcss.replace(oldsel, oldsel + ', ' + newsel)
        assert s.cssText == expcss.encode()
        # Selector stylerule
        with pytest.raises(xml.dom.NamespaceErr):
            sr.selectorList.append('x|b')
        oldsel, newsel = sr.selectorText, 'ex2|SELSR'
        sr.selectorList.append(newsel)
        expcss = expcss.replace(oldsel, oldsel + ', ' + newsel)
        assert s.cssText == expcss.encode()

        assert (
            s.cssText
            == b'''@namespace n "new";
@namespace ex2 "example";
@namespace DEFAULT "default";
@media all {
    n|SEL3, a, ex2|SELMR {
        left: 0
        }
    }
ex2|SEL4, a, ex2|SELSR {
    top: 0
    }'''
        )

    def test_namespaces3(self):
        "CSSStyleSheet.namespaces 3"
        # tests setting namespaces with new {}
        s = cssutils.css.CSSStyleSheet()
        css = 'h|a { top: 0 }'
        with pytest.raises(xml.dom.NamespaceErr):
            s.add(css)

        s.add('@namespace x "html";')
        assert s.namespaces['x'] == 'html'

        r = cssutils.css.CSSStyleRule()
        r.cssText = (css, {'h': 'html'})
        s.add(r)  # uses x as set before! h is temporary only
        assert s.cssText == b'@namespace x "html";\nx|a {\n    top: 0\n    }'

        # prefix is now "x"!
        with pytest.raises(xml.dom.NamespaceErr):
            r.selectorList.append('h|b')
        with pytest.raises(xml.dom.NamespaceErr):
            r.selectorList.append('y|b')
        s.namespaces['y'] = 'html'
        r.selectorList.append('y|b')
        assert s.cssText == b'@namespace y "html";\ny|a, y|b {\n    top: 0\n    }'

        with pytest.raises(xml.dom.NoModificationAllowedErr):
            s.namespaces.__delitem__('y')
        assert s.cssText == b'@namespace y "html";\ny|a, y|b {\n    top: 0\n    }'

        s.cssRules[0].prefix = ''
        assert s.cssText == b'@namespace "html";\na, b {\n    top: 0\n    }'

        # remove need of namespace
        s.cssRules[0].prefix = 'x'
        s.cssRules[1].selectorText = 'a, b'
        assert s.cssText == b'@namespace x "html";\na, b {\n    top: 0\n    }'

    def test_namespaces4(self):
        "CSSStyleSheet.namespaces 4"
        # tests setting namespaces with new {}
        s = cssutils.css.CSSStyleSheet()
        assert {} == s.namespaces.namespaces

        s.namespaces.namespaces['a'] = 'no setting possible'
        assert {} == s.namespaces.namespaces

        s.namespaces[None] = 'default'
        assert {'': 'default'} == s.namespaces.namespaces

        del s.namespaces['']
        assert {} == s.namespaces.namespaces

        s.namespaces[''] = 'default'
        assert {'': 'default'} == s.namespaces.namespaces

        del s.namespaces[None]
        assert {} == s.namespaces.namespaces

        s.namespaces['p'] = 'uri'
        # cannot use namespaces.namespaces
        del s.namespaces.namespaces['p']
        assert {'p': 'uri'} == s.namespaces.namespaces

        with pytest.raises(xml.dom.NamespaceErr, match=r"Prefix undefined not found\."):
            del s.namespaces['undefined']

    def test_namespaces5(self):
        "CSSStyleSheet.namespaces 5"
        # unknown namespace
        s = cssutils.parseString('a|a { color: red }')
        assert s.cssText == b''

        s = cssutils.css.CSSStyleSheet()
        with pytest.raises(xml.dom.NamespaceErr, match=r"Prefix a not found\."):
            s._setCssText('a|a { color: red }')

    def test_deleteRuleIndex(self):
        "CSSStyleSheet.deleteRule(index)"
        self.s.cssText = '@charset "ascii"; @import "x"; @x; a {\n    x: 1\n    }@y;'
        assert 5 == self.s.cssRules.length

        with pytest.raises(xml.dom.IndexSizeErr):
            self.s.deleteRule(5)

        # end -1
        # check parentStyleSheet
        r = self.s.cssRules[-1]
        assert self.s == r.parentStyleSheet
        self.s.deleteRule(-1)
        assert r.parentStyleSheet is None

        assert 4 == self.s.cssRules.length
        assert (
            b'@charset "ascii";\n@import "x";\n@x;\na {\n    x: 1\n    }'
            == self.s.cssText
        )
        # beginning
        self.s.deleteRule(0)
        assert 3 == self.s.cssRules.length
        assert b'@import "x";\n@x;\na {\n    x: 1\n    }' == self.s.cssText
        # middle
        self.s.deleteRule(1)
        assert 2 == self.s.cssRules.length
        assert b'@import "x";\na {\n    x: 1\n    }' == self.s.cssText
        # end
        self.s.deleteRule(1)
        assert 1 == self.s.cssRules.length
        assert b'@import "x";' == self.s.cssText

    def test_deleteRule(self):
        "CSSStyleSheet.deleteRule(rule)"
        s = cssutils.css.CSSStyleSheet()
        s.cssText = '''
        @namespace x "http://example.com";
        a { color: red; }
        b { color: blue; }
        c { color: green; }
        '''
        n, s1, s2, s3 = s.cssRules

        r = cssutils.css.CSSStyleRule()
        with pytest.raises(xml.dom.IndexSizeErr):
            s.deleteRule(r)

        assert 4 == s.cssRules.length
        s.deleteRule(n)
        assert 3 == s.cssRules.length
        assert (
            s.cssText
            == b'a {\n    color: red\n    }\nb {\n    color: blue\n    }\nc {\n    color: green\n    }'
        )
        with pytest.raises(xml.dom.IndexSizeErr):
            s.deleteRule(n)
        s.deleteRule(s2)
        assert 2 == s.cssRules.length
        assert s.cssText == b'a {\n    color: red\n    }\nc {\n    color: green\n    }'
        with pytest.raises(xml.dom.IndexSizeErr):
            s.deleteRule(s2)

    def _gets(self):
        # complete
        self.cr = cssutils.css.CSSCharsetRule('ascii')
        self.c = cssutils.css.CSSComment('/*c*/')
        self.ur = cssutils.css.CSSUnknownRule('@x;')
        self.ir = cssutils.css.CSSImportRule('x')
        self.nr = cssutils.css.CSSNamespaceRule('uri', 'p')
        self.mr = cssutils.css.CSSMediaRule()
        self.mr.cssText = '@media all { @m; }'
        self.pr = cssutils.css.CSSPageRule()
        self.pr.style = 'margin: 0;'
        self.sr = cssutils.css.CSSStyleRule()
        self.sr.cssText = 'a {\n    x: 1\n    }'

        s = cssutils.css.CSSStyleSheet()
        s.insertRule(self.cr)  # 0
        s.insertRule(self.ir)  # 1
        s.insertRule(self.nr)  # 2
        s.insertRule(self.mr)  # 3
        s.insertRule(self.sr)  # 4
        s.insertRule(self.mr)  # 5
        s.insertRule(self.pr)  # 6
        s.insertRule(self.sr)  # 7
        assert (
            b'@charset "ascii";\n@import url(x);\n@namespace p "uri";\n@media all {\n    @m;\n    }\na {\n    x: 1\n    }\n@media all {\n    @m;\n    }\n@page {\n    margin: 0\n    }\na {\n    x: 1\n    }'
            == s.cssText
        )
        return s, s.cssRules.length

    def test_add(self):
        "CSSStyleSheet.add()"
        full = cssutils.css.CSSStyleSheet()
        sheet = cssutils.css.CSSStyleSheet()
        css = [
            '@charset "ascii";',
            '@import "x";',
            '@namespace p "u";',
            '@page {\n    left: 0\n    }',
            '@font-face {\n    src: local(x)\n    }',
            '@media all {\n    a {\n        color: red\n        }\n    }',
            'a {\n    color: green\n    }',
            '/*comment*/',
            '@x;',
        ]

        fullcss = '\n'.join(css)
        full.cssText = fullcss
        assert full.cssText == fullcss.encode()
        for i, line in enumerate(css):
            # sheet without same ruletype
            before = css[:i]
            after = css[i + 1 :]
            sheet.cssText = ''.join(before + after)

            index = sheet.add(line)
            if i < 3:
                # specific insertion point
                assert fullcss.encode() == sheet.cssText
                assert i == index
            else:
                # end of sheet
                expected = before
                expected.extend(after)
                expected.append(line)
                assert '\n'.join(expected).encode() == sheet.cssText
                assert len(expected) - 1 == index  # no same rule present

            # sheet with the same ruletype
            if i == 1:
                line = '@import "x2";'
            if i == 2:
                line = '@namespace p2 "u2";'

            full.cssText = fullcss
            index = full.add(line)
            if i < 1:
                assert fullcss.encode() == sheet.cssText
                assert i == index  # no same rule present
            else:
                if i < 3:
                    # in order
                    expected = css[: i + 1]  # including same rule
                    expected.append(line)
                    expected.extend(css[i + 1 :])
                    expectedindex = i + 1
                else:
                    # just appended as no order needed
                    expected = css[:]
                    expected.append(line)
                    expectedindex = len(expected) - 1

                assert '\n'.join(expected).encode() == full.cssText
                assert expectedindex == index  # no same rule present

    def test_addimport(self):
        p = cssutils.CSSParser(fetcher=lambda url: (None, '/**/'))

        cssrulessheet = p.parseString('@import "example.css";')
        imports = (
            '@import "example.css";',  # string
            cssutils.css.CSSImportRule(href="example.css"),  # CSSRule
            cssrulessheet.cssRules,  # CSSRuleList
        )
        for imp in imports:
            sheet = p.parseString('', href='http://example.com')
            sheet.add(imp)
            added = sheet.cssRules[0]
            assert sheet == added.parentStyleSheet
            assert 'example.css' == added.href
            assert 'utf-8' == added.styleSheet.encoding
            assert b'/**/' == added.styleSheet.cssText

        cssrulessheet = p.parseString('@import "example.css";')
        imports = (
            ('@import "example.css";', 'ascii'),  # string
            (cssutils.css.CSSImportRule(href="example.css"), 'ascii'),  # CSSRule
            # got encoding from old parent already
            (cssrulessheet.cssRules, 'utf-8'),  # CSSRuleList
        )
        for imp, enc in imports:
            sheet = p.parseString('', href='http://example.com', encoding='ascii')
            sheet.add(imp)
            added = sheet.cssRules[1]
            assert sheet == added.parentStyleSheet
            assert 'example.css' == added.href
            assert enc == added.styleSheet.encoding
            if enc == 'ascii':
                assert b'@charset "ascii";\n/**/' == added.styleSheet.cssText
            else:
                assert b'/**/' == added.styleSheet.cssText

        # if styleSheet is already there encoding is not set new
        impsheet = p.parseString('@import "example.css";')
        imp = impsheet.cssRules[0]
        sheet = p.parseString('', href='http://example.com', encoding='ascii')
        sheet.add(imp)
        added = sheet.cssRules[1]
        assert sheet == added.parentStyleSheet
        assert 'example.css' == added.href
        assert 'utf-8' == added.styleSheet.encoding
        assert b'/**/' == added.styleSheet.cssText

    def test_insertRule(self):
        "CSSStyleSheet.insertRule()"
        s, L = self._gets()

        # INVALID index
        with pytest.raises(xml.dom.IndexSizeErr):
            s.insertRule(self.sr, -1)
        with pytest.raises(xml.dom.IndexSizeErr):
            s.insertRule(self.sr, s.cssRules.length + 1)
        #   check if rule is really not in
        assert L == s.cssRules.length

        # insert string
        s.insertRule('a {}')
        assert L + 1 == s.cssRules.length
        # insert rule
        s.insertRule(self.sr)
        assert L + 2 == s.cssRules.length
        # insert rulelist
        s2, L2 = self._gets()
        rulelist = s2.cssRules
        del rulelist[:-2]
        s.insertRule(rulelist)
        assert L + 2 + 2 == s.cssRules.length

    def _insertRule(self, rules, notbefore, notafter, anywhere, checkdoubles=True):
        """
        helper
        test if any rule in rules cannot be inserted before rules in before
        or after rules in after but before and after rules in anywhere
        """
        for rule in rules:
            for r in notbefore:
                s = cssutils.css.CSSStyleSheet()
                s.insertRule(r)
                with pytest.raises(xml.dom.HierarchyRequestErr):
                    s.insertRule(rule, 0)
                s = cssutils.css.CSSStyleSheet()
                s.add(r)
                with pytest.raises(xml.dom.HierarchyRequestErr):
                    s.insertRule(rule, 0)
            for r in notafter:
                s = cssutils.css.CSSStyleSheet()
                s.insertRule(r)
                with pytest.raises(xml.dom.HierarchyRequestErr):
                    s.insertRule(rule, 1)
                s = cssutils.css.CSSStyleSheet()
                s.add(r)
                s.add(rule)  # never raises
                assert s == r.parentStyleSheet

            for r in anywhere:
                s = cssutils.css.CSSStyleSheet()
                s.insertRule(r)
                s.insertRule(rule, 0)  # before
                s.insertRule(rule)  # after
                if checkdoubles:
                    assert s.cssRules.length == 3
                assert s == r.parentStyleSheet

    def test_insertRule_charset(self):
        "CSSStyleSheet.insertRule(@charset)"
        s, L = self._gets()
        notbefore = (self.cr,)
        notafter = (
            self.cr,
            self.ir,
            self.nr,
            self.mr,
            self.pr,
            self.sr,
            self.c,
            self.ur,
        )
        anywhere = ()
        self._insertRule((self.cr,), notbefore, notafter, anywhere)

    def test_insertRule_import(self):
        "CSSStyleSheet.insertRule(@import)"
        s, L = self._gets()
        notbefore = (self.cr,)
        notafter = (self.nr, self.pr, self.mr, self.sr)
        anywhere = (self.c, self.ur, self.ir)
        self._insertRule((self.ir,), notbefore, notafter, anywhere)

    def test_insertRule_namespace(self):
        "CSSStyleSheet.insertRule(@namespace)"
        s, L = self._gets()
        notbefore = (self.cr, self.ir)
        notafter = (self.pr, self.mr, self.sr)
        anywhere = (self.c, self.ur, self.nr)
        self._insertRule((self.nr,), notbefore, notafter, anywhere, checkdoubles=False)

    def test_insertRule_media_page_style(self):
        "CSSStyleSheet.insertRule(@media, @page, stylerule)"
        s, L = self._gets()
        notbefore = (self.cr, self.ir, self.nr)
        notafter = ()
        anywhere = (self.c, self.ur, self.mr, self.pr, self.sr)
        self._insertRule((self.pr, self.mr, self.sr), notbefore, notafter, anywhere)

    def test_insertRule_unknownandcomment(self):
        "CSSStyleSheet.insertRule(@ unknown, comment)"
        s, L = self._gets()
        notbefore = (self.cr,)
        notafter = ()
        anywhere = (self.c, self.ur, self.ir, self.nr, self.mr, self.pr, self.sr)
        self._insertRule((self.ur,), notbefore, notafter, anywhere)

    def test_HTMLComments(self):
        "CSSStyleSheet CDO CDC"
        css = '''body { color: red }
<!-- comment -->
body { color: blue }
body { color: pink }
<!-- comment -->
body { color: green }
'''
        exp = '''body {
    color: red
    }
body {
    color: pink
    }'''
        sheet = cssutils.parseString(css)
        assert sheet.cssText == exp.encode()

    def test_incomplete(self):
        "CSSStyleRule (incomplete)"
        tests = {
            '@import "a': '@import "a";',  # no }
            'a { x: 1': 'a {\n    x: 1\n    }',  # no }
            'a { font-family: "arial sans': 'a {\n    font-family: "arial sans"\n    }',  # no "
        }
        self.do_equal_p(tests)  # parse

    def test_NoModificationAllowedErr(self):
        "CSSStyleSheet NoModificationAllowedErr"
        css = cssutils.css.CSSStyleSheet(readonly=True)

        assert css._readonly  # internal...

        with pytest.raises(xml.dom.NoModificationAllowedErr):
            css._setCssText('@x;')
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            css.insertRule(self.rule)
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            css.insertRule(self.rule, 0)
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            css.deleteRule(0)

    def test_reprANDstr(self):
        "CSSStyleSheet.__repr__(), .__str__()"
        href = 'file:foo.css'
        title = 'title-of-css'

        s = cssutils.css.CSSStyleSheet(href=href, title=title)

        assert href in str(s)
        assert title in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert href == s2.href
        assert title == s2.title

    def test_valid(self):
        cases = [
            ('body { color: red; }', True),
            ('body { color: asd; }', False),
            ('body { foo: 12px; }', False),
        ]
        for case, expected in cases:
            sheet = cssutils.parseString(case)
            msg = "{!r} should be {}".format(case, 'valid' if expected else 'invalid')
            assert sheet.valid == expected, msg



"""Testcases for cssutils.css.CSSMediaRule"""

# flake8: noqa

import xml.dom
from . import test_cssrule
import cssutils
import pytest


class TestCSSMediaRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSMediaRule()
        self.rRO = cssutils.css.CSSMediaRule(readonly=True)
        self.r_type = cssutils.css.CSSMediaRule.MEDIA_RULE
        self.r_typeString = 'MEDIA_RULE'
        # for tests
        self.stylerule = cssutils.css.CSSStyleRule()
        self.stylerule.cssText = 'a {}'

    def test_init(self):
        "CSSMediaRule.__init__()"
        super().test_init()

        r = cssutils.css.CSSMediaRule()
        assert isinstance(r.cssRules, cssutils.css.CSSRuleList)
        assert [] == r.cssRules
        assert '' == r.cssText
        assert isinstance(r.media, cssutils.stylesheets.MediaList)
        assert 'all' == r.media.mediaText
        assert r.name is None

        r = cssutils.css.CSSMediaRule(mediaText='print', name='name')
        assert isinstance(r.cssRules, cssutils.css.CSSRuleList)
        assert [] == r.cssRules
        assert '' == r.cssText
        assert isinstance(r.media, cssutils.stylesheets.MediaList)
        assert 'print' == r.media.mediaText
        assert 'name' == r.name

        # only possible to set @... similar name
        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setAtkeyword('x')

    def test_iter(self):
        "CSSMediaRule.__iter__()"
        m = cssutils.css.CSSMediaRule()
        m.cssText = '''@media all { /*1*/a { left: 0} b{ top:0} }'''
        types = [
            cssutils.css.CSSRule.COMMENT,
            cssutils.css.CSSRule.STYLE_RULE,
            cssutils.css.CSSRule.STYLE_RULE,
        ]
        for i, rule in enumerate(m):
            assert rule == m.cssRules[i]
            assert rule.type == types[i]
            assert rule.parentRule == m

    def test_refs(self):
        """CSSStylesheet references"""
        s = cssutils.parseString('@media all {a {color: red}}')
        r = s.cssRules[0]
        rules = r.cssRules
        assert r.cssRules[0].parentStyleSheet == s
        assert rules[0].parentStyleSheet == s

        # set cssText
        r.cssText = '@media all {a {color: blue}}'
        # not anymore: self.assertEqual(rules, r.cssRules)

        # set cssRules
        r.cssRules = cssutils.parseString(
            '''
            /**/
            @x;
            b {}').cssRules'''
        ).cssRules
        # new object
        assert rules != r.cssRules
        for i, sr in enumerate(r.cssRules):
            assert sr.parentStyleSheet == s
            assert sr.parentRule == r

    def test_cssRules(self):
        "CSSMediaRule.cssRules"
        r = cssutils.css.CSSMediaRule()
        assert [] == r.cssRules
        sr = cssutils.css.CSSStyleRule()
        r.cssRules.append(sr)
        assert [sr] == r.cssRules
        ir = cssutils.css.CSSImportRule()
        with pytest.raises(xml.dom.HierarchyRequestErr):
            r.cssRules.append(ir)

        s = cssutils.parseString('@media all { /*1*/a {x:1} }')
        m = s.cssRules[0]
        assert 2 == m.cssRules.length
        del m.cssRules[0]
        assert 1 == m.cssRules.length
        m.cssRules.append('/*2*/')
        assert 2 == m.cssRules.length
        m.cssRules.extend(cssutils.parseString('/*3*/x {y:2}').cssRules)
        assert 4 == m.cssRules.length
        assert (
            '@media all {\n    a {\n        x: 1\n        }\n    /*2*/\n    /*3*/\n    x {\n        y: 2\n        }\n    }'
            == m.cssText
        )

        for rule in m.cssRules:
            assert rule.parentStyleSheet == s
            assert rule.parentRule == m

    def test_cssText(self):
        "CSSMediaRule.cssText"
        style = '''{
    a {
        color: red
        }
    }'''

        mls = {
            ' (min-device-pixel-ratio: 1.3), (min-resolution: 1.3dppx) ': None,
            ' tv ': None,
            ' only tv ': None,
            ' not tv ': None,
            ' only tv and (color) ': None,
            ' only tv and(color)': ' only tv and (color) ',
            ' only tv and (color: red) ': None,
            ' only tv and (color: red) and (width: 100px) ': None,
            ' only tv and (color: red) and (width: 100px), tv ': None,
            ' only tv and (color: red) and (width: 100px), tv and (width: 20px) ': None,
            ' only tv and(color :red)and(  width :100px  )  ,tv and(width: 20px) ': ' only tv and (color: red) and (width: 100px), tv and (width: 20px) ',
            ' (color: red) and (width: 100px), (width: 20px) ': None,
            ' /*1*/ only /*2*/ tv /*3*/ and /*4*/ (/*5*/ width) /*5*/ /*6*/, (color) and (height) ': None,
            '(color)and(width),(height)': ' (color) and (width), (height) ',
        }
        tests = {}
        for b, a in list(mls.items()):
            if a is None:
                a = b
            tests[f'@media{b}{style}'] = f'@media{a}{style}'

        self.do_equal_p(tests)
        self.do_equal_r(tests)

        tests = {
            '@media only tv{}': '',
            '@media not tv{}': '',
            '@media only tv and (color){}': '',
            '@media only tv and (color: red){}': '',
            '@media only tv and (color: red) and (width: 100px){}': '',
            '@media only tv and (color: red) and (width: 100px), tv{}': '',
            '@media only tv and (color: red) and (width: 100px), tv and (width: 20px){}': '',
            '@media (color: red) and (width: 100px), (width: 20px){}': '',
            '@media (width){}': '',
            '@media (width:10px){}': '',
            '@media (width), (color){}': '',
            '@media (width)  ,  (color),(height){}': '',
            '@media (width)  ,  (color) and (height){}': '',
            '@media (width) and (color){}': '',
            '@media all and (width){}': '',
            '@media all and (width:10px){}': '',
            '@media all and (width), (color){}': '',
            '@media all and (width)  ,  (color),(height){}': '',
            '@media all and (width)  ,  (color) and (height){}': '',
            '@media all and (width) and (color){}': '',
            '@media only tv and (width){}': '',
            '@media only tv and (width:10px){}': '',
            '@media only tv and (width), (color){}': '',
            '@media only tv and (width)  ,  (color),(height){}': '',
            '@media only tv and (width)  ,  (color) and (height){}': '',
            '@media only tv and (width) and (color){}': '',
            '@media only tv and (width) "name" {}': '',
            '@media only tv and (width:10px) "name" {}': '',
            '@media only tv and (width), (color){}': '',
            '@media only tv and (width)  ,  (color),(height){}': '',
            '@media only tv and (width)  ,  (color) and (height){}': '',
            '@media only tv and (width) and (color){}': '',
            '@media all "name"{}': '',
            '@media all {}': '',
            '@media/*x*/all{}': '',
            '@media all { a{ x: 1} }': '@media all {\n    a {\n        x: 1\n        }\n    }',
            '@media all "name" { a{ x: 1} }': '@media all "name" {\n    a {\n        x: 1\n        }\n    }',
            '@MEDIA all { a{x:1} }': '@media all {\n    a {\n        x: 1\n        }\n    }',
            '@\\media all { a{x:1} }': '@media all {\n    a {\n        x: 1\n        }\n    }',
            '@media all {@x some;a{color: red;}b{color: green;}}': '''@media all {
    @x some;
    a {
        color: red
        }
    b {
        color: green
        }
    }''',
            '@media all { @x{}}': '@media all {\n    @x {\n        }\n    }',
            '@media all "n" /**/ { @x{}}': '@media all "n" /**/ {\n    @x {\n        }\n    }',
            # comments
            '@media/*1*//*2*/all/*3*//*4*/{/*5*/a{x:1}}': '@media /*1*/ /*2*/ all /*3*/ /*4*/ {\n    /*5*/\n    a {\n        x: 1\n        }\n    }',
            '@media  /*1*/  /*2*/  all  /*3*/  /*4*/  {  /*5*/  a{ x: 1} }': '@media /*1*/ /*2*/ all /*3*/ /*4*/ {\n    /*5*/\n    a {\n        x: 1\n        }\n    }',
            # WS
            '@media\n\t\f all\n\t\f {\n\t\f a{ x: 1}\n\t\f }': '@media all {\n    a {\n        x: 1\n        }\n    }',
            # @page rule inside @media
            '@media all { @page { margin: 0; } }': '@media all {\n    @page {\n        margin: 0\n        }\n    }',
            # nested media rules
            '@media all { @media all { p { color: red; } } }': '@media all {\n    @media all {\n        p {\n            '
            'color: red\n            }\n        }\n    }',
        }
        self.do_equal_p(tests)
        self.do_equal_r(tests)

        tests = {
            '@media {}': xml.dom.SyntaxErr,
            '@media;': xml.dom.SyntaxErr,
            '@media/*only comment*/{}': xml.dom.SyntaxErr,
            '@media all;': xml.dom.SyntaxErr,
            '@media all "n";': xml.dom.SyntaxErr,
            '@media all; @x{}': xml.dom.SyntaxErr,
            '@media { a{ x: 1} }': xml.dom.SyntaxErr,
            '@media "name" { a{ x: 1} }': xml.dom.SyntaxErr,
            '@media "name" all { a{ x: 1} }': xml.dom.SyntaxErr,
            '@media all { @charset "x"; a{}}': xml.dom.HierarchyRequestErr,
            '@media all { @import "x"; a{}}': xml.dom.HierarchyRequestErr,
            '@media all { , }': xml.dom.SyntaxErr,
            '@media all {}EXTRA': xml.dom.SyntaxErr,
            '@media ({}': xml.dom.SyntaxErr,
            '@media (color{}': xml.dom.SyntaxErr,
            '@media (color:{}': xml.dom.SyntaxErr,
            '@media (color:red{}': xml.dom.SyntaxErr,
            '@media (:red){}': xml.dom.SyntaxErr,
            '@media (:){}': xml.dom.SyntaxErr,
            '@media color:red){}': xml.dom.SyntaxErr,
        }
        self.do_raise_p(tests)
        self.do_raise_r(tests)

        tests = {
            # extra stuff
            '@media all { x{} } a{}': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests)

        m = cssutils.css.CSSMediaRule()
        m.cssText = '''@media all {@x; /*1*/a{color: red;}}'''
        for r in m.cssRules:
            assert m == r.parentRule
            assert m.parentStyleSheet == r.parentStyleSheet

    def test_media(self):
        "CSSMediaRule.media"
        # see CSSImportRule.media

        # setting not allowed
        with pytest.raises(AttributeError):
            self.r.__setattr__('media', None)
        with pytest.raises(AttributeError):
            self.r.__setattr__('media', 0)

        # set mediaText instead
        self.r.media.mediaText = 'print'
        self.r.insertRule(self.stylerule)
        assert '' == self.r.cssText
        cssutils.ser.prefs.keepEmptyRules = True
        assert '@media print {\n    a {}\n    }' == self.r.cssText

    def test_name(self):
        "CSSMediaRule.name"
        r = cssutils.css.CSSMediaRule()
        r.cssText = '@media all "\\n\\"ame" {a{left: 0}}'

        assert '\\n"ame' == r.name
        r.name = "n"
        assert 'n' == r.name
        assert (
            '@media all "n" {\n    a {\n        left: 0\n        }\n    }' == r.cssText
        )
        r.name = '"'
        assert '"' == r.name
        assert (
            '@media all "\\"" {\n    a {\n        left: 0\n        }\n    }'
            == r.cssText
        )

        r.name = ''
        assert r.name is None
        assert '@media all {\n    a {\n        left: 0\n        }\n    }' == r.cssText

        r.name = None
        assert r.name is None
        assert '@media all {\n    a {\n        left: 0\n        }\n    }' == r.cssText

        with pytest.raises(xml.dom.SyntaxErr):
            r._setName(0)
        with pytest.raises(xml.dom.SyntaxErr):
            r._setName(123)

    def test_deleteRuleIndex(self):
        "CSSMediaRule.deleteRule(index)"
        # see CSSStyleSheet.deleteRule
        m = cssutils.css.CSSMediaRule()
        m.cssText = '''@media all {
            @a;
            /* x */
            @b;
            @c;
            @d;
        }'''
        assert 5 == m.cssRules.length
        with pytest.raises(xml.dom.IndexSizeErr):
            m.deleteRule(5)

        # end -1
        # check parentRule
        r = m.cssRules[-1]
        assert m == r.parentRule
        m.deleteRule(-1)
        assert r.parentRule is None

        assert 4 == m.cssRules.length
        assert (
            '@media all {\n    @a;\n    /* x */\n    @b;\n    @c;\n    }' == m.cssText
        )
        # beginning
        m.deleteRule(0)
        assert 3 == m.cssRules.length
        assert '@media all {\n    /* x */\n    @b;\n    @c;\n    }' == m.cssText
        # middle
        m.deleteRule(1)
        assert 2 == m.cssRules.length
        assert '@media all {\n    /* x */\n    @c;\n    }' == m.cssText
        # end
        m.deleteRule(1)
        assert 1 == m.cssRules.length
        assert '@media all {\n    /* x */\n    }' == m.cssText

    def test_deleteRule(self):
        "CSSMediaRule.deleteRule(rule)"
        m = cssutils.css.CSSMediaRule()
        m.cssText = '''@media all {
            a { color: red; }
            b { color: blue; }
            c { color: green; }
        }'''
        s1, s2, s3 = m.cssRules

        r = cssutils.css.CSSStyleRule()
        with pytest.raises(xml.dom.IndexSizeErr):
            m.deleteRule(r)

        assert 3 == m.cssRules.length
        m.deleteRule(s2)
        assert 2 == m.cssRules.length
        assert (
            m.cssText
            == '@media all {\n    a {\n        color: red\n        }\n    c {\n        color: green\n        }\n    }'
        )
        with pytest.raises(xml.dom.IndexSizeErr):
            m.deleteRule(s2)

    def test_add(self):
        "CSSMediaRule.add()"
        # see CSSStyleSheet.add
        r = cssutils.css.CSSMediaRule()
        stylerule1 = cssutils.css.CSSStyleRule()
        stylerule2 = cssutils.css.CSSStyleRule()
        r.add(stylerule1)
        r.add(stylerule2)
        assert r.cssRules[0] == stylerule1
        assert r.cssRules[1] == stylerule2

    def test_insertRule(self):
        "CSSMediaRule.insertRule"
        # see CSSStyleSheet.insertRule
        r = cssutils.css.CSSMediaRule()
        charsetrule = cssutils.css.CSSCharsetRule('ascii')
        importrule = cssutils.css.CSSImportRule('x')
        namespacerule = cssutils.css.CSSNamespaceRule()
        unknownrule = cssutils.css.CSSUnknownRule('@x;')
        stylerule = cssutils.css.CSSStyleRule('a')
        stylerule.cssText = 'a { x: 1}'
        comment1 = cssutils.css.CSSComment('/*1*/')
        comment2 = cssutils.css.CSSComment('/*2*/')

        # hierarchy
        with pytest.raises(xml.dom.HierarchyRequestErr):
            r.insertRule(charsetrule, 0)
        with pytest.raises(xml.dom.HierarchyRequestErr):
            r.insertRule(importrule, 0)
        with pytest.raises(xml.dom.HierarchyRequestErr):
            r.insertRule(namespacerule, 0)

        # start insert
        r.insertRule(stylerule, 0)
        assert r == stylerule.parentRule
        assert r.parentStyleSheet == stylerule.parentStyleSheet
        # before
        r.insertRule(comment1, 0)
        assert r == comment1.parentRule
        assert r.parentStyleSheet == stylerule.parentStyleSheet
        # end explicit
        r.insertRule(unknownrule, 2)
        assert r == unknownrule.parentRule
        assert r.parentStyleSheet == stylerule.parentStyleSheet
        # end implicit
        r.insertRule(comment2)
        assert r == comment2.parentRule
        assert r.parentStyleSheet == stylerule.parentStyleSheet
        assert (
            '@media all {\n    /*1*/\n    a {\n        x: 1\n        }\n    @x;\n    /*2*/\n    }'
            == r.cssText
        )

        # index
        with pytest.raises(xml.dom.IndexSizeErr):
            r.insertRule(stylerule, -1)
        with pytest.raises(xml.dom.IndexSizeErr):
            r.insertRule(stylerule, r.cssRules.length + 1)

    def test_InvalidModificationErr(self):
        "CSSMediaRule.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('@media')

    def test_incomplete(self):
        "CSSMediaRule (incomplete)"
        tests = {
            '@media all { @unknown;': '@media all {\n    @unknown;\n    }',  # no }
            '@media all { a {x:"1"}': '@media all {\n    a {\n        x: "1"\n        }\n    }',  # no }
            '@media all { a {x:"1"': '@media all {\n    a {\n        x: "1"\n        }\n    }',  # no }}
            '@media all { a {x:"1': '@media all {\n    a {\n        x: "1"\n        }\n    }',  # no "}}
        }
        self.do_equal_p(tests)  # parse

    def test_reprANDstr(self):
        "CSSMediaRule.__repr__(), .__str__()"
        mediaText = 'tv, print'

        s = cssutils.css.CSSMediaRule(mediaText=mediaText)

        assert mediaText in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert mediaText == s2.media.mediaText



"""CSSComment is not defined in DOM Level 2 at all but a cssutils defined
class only.

Implements CSSRule which is also extended for a CSSComment rule type.
"""

__all__ = ['CSSComment']

import xml.dom

import cssutils

from . import cssrule


class CSSComment(cssrule.CSSRule):
    """
    Represents a CSS comment (cssutils only).

    Format::

        /*...*/
    """

    def __init__(
        self, cssText=None, parentRule=None, parentStyleSheet=None, readonly=False
    ):
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)

        self._cssText = None
        if cssText:
            self._setCssText(cssText)

        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.css.{self.__class__.__name__}(cssText={self.cssText!r})"

    def __str__(self):
        return f"<cssutils.css.{self.__class__.__name__} object cssText={self.cssText!r} at 0x{id(self):x}>"

    def _getCssText(self):
        """Return serialized property cssText."""
        return cssutils.ser.do_CSSComment(self)

    def _setCssText(self, cssText):
        """
        :param cssText:
            textual text to set or tokenlist which is not tokenized
            anymore. May also be a single token for this rule

        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
        """
        super()._setCssText(cssText)
        tokenizer = self._tokenize2(cssText)

        commenttoken = self._nexttoken(tokenizer)
        unexpected = self._nexttoken(tokenizer)

        if (
            not commenttoken
            or self._type(commenttoken) != self._prods.COMMENT
            or unexpected
        ):
            self._log.error(
                'CSSComment: Not a CSSComment: %r' % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            self._cssText = self._tokenvalue(commenttoken)

    cssText = property(
        _getCssText,
        _setCssText,
        doc="The parsable textual representation of this rule.",
    )

    type = property(
        lambda self: self.COMMENT,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    # constant but needed:
    wellformed = property(lambda self: True)



"""CSSUnknownRule implements DOM Level 2 CSS CSSUnknownRule."""

__all__ = ['CSSUnknownRule']

import xml.dom

import cssutils

from . import cssrule


class CSSUnknownRule(cssrule.CSSRule):
    """
    Represents an at-rule not supported by this user agent, so in
    effect all other at-rules not defined in cssutils.

    Format::

        @xxx until ';' or block {...}
    """

    def __init__(
        self, cssText='', parentRule=None, parentStyleSheet=None, readonly=False
    ):
        """
        :param cssText:
            of type string
        """
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)
        self._atkeyword = None
        if cssText:
            self.cssText = cssText

        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.css.{self.__class__.__name__}(cssText={self.cssText!r})"

    def __str__(self):
        return f"<cssutils.css.{self.__class__.__name__} object cssText={self.cssText!r} at 0x{id(self):x}>"

    def _getCssText(self):
        """Return serialized property cssText."""
        return cssutils.ser.do_CSSUnknownRule(self)

    def _setCssText(self, cssText):  # noqa: C901
        """
        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
        """
        super()._setCssText(cssText)
        tokenizer = self._tokenize2(cssText)
        attoken = self._nexttoken(tokenizer, None)
        if not attoken or self._type(attoken) != self._prods.ATKEYWORD:
            self._log.error(
                'CSSUnknownRule: No CSSUnknownRule found: %s' % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            # for closures: must be a mutable
            new = {'nesting': [], 'wellformed': True}  # {} [] or ()

            def CHAR(expected, seq, token, tokenizer=None):
                type_, val, line, col = token
                if expected != 'EOF':
                    if val in '{[(':
                        new['nesting'].append(val)
                    elif val in '}])':
                        opening = {'}': '{', ']': '[', ')': '('}[val]
                        try:
                            if new['nesting'][-1] == opening:
                                new['nesting'].pop()
                            else:
                                raise IndexError()
                        except IndexError:
                            new['wellformed'] = False
                            self._log.error(
                                'CSSUnknownRule: Wrong nesting of ' '{, [ or (.',
                                token=token,
                            )

                    if val in '};' and not new['nesting']:
                        expected = 'EOF'

                    seq.append(val, type_, line=line, col=col)
                    return expected
                else:
                    new['wellformed'] = False
                    self._log.error(
                        'CSSUnknownRule: Expected end of rule.', token=token
                    )
                    return expected

            def FUNCTION(expected, seq, token, tokenizer=None):
                # handled as opening (
                type_, val, line, col = token
                val = self._tokenvalue(token)
                if expected != 'EOF':
                    new['nesting'].append('(')
                    seq.append(val, type_, line=line, col=col)
                    return expected
                else:
                    new['wellformed'] = False
                    self._log.error(
                        'CSSUnknownRule: Expected end of rule.', token=token
                    )
                    return expected

            def EOF(expected, seq, token, tokenizer=None):
                "close all blocks and return 'EOF'"
                for x in reversed(new['nesting']):
                    closing = {'{': '}', '[': ']', '(': ')'}[x]
                    seq.append(closing, closing)
                new['nesting'] = []
                return 'EOF'

            def INVALID(expected, seq, token, tokenizer=None):
                # makes rule invalid
                self._log.error(
                    'CSSUnknownRule: Bad syntax.', token=token, error=xml.dom.SyntaxErr
                )
                new['wellformed'] = False
                return expected

            def STRING(expected, seq, token, tokenizer=None):
                type_, val, line, col = token
                val = self._stringtokenvalue(token)
                if expected != 'EOF':
                    seq.append(val, type_, line=line, col=col)
                    return expected
                else:
                    new['wellformed'] = False
                    self._log.error(
                        'CSSUnknownRule: Expected end of rule.', token=token
                    )
                    return expected

            def URI(expected, seq, token, tokenizer=None):
                type_, val, line, col = token
                val = self._uritokenvalue(token)
                if expected != 'EOF':
                    seq.append(val, type_, line=line, col=col)
                    return expected
                else:
                    new['wellformed'] = False
                    self._log.error(
                        'CSSUnknownRule: Expected end of rule.', token=token
                    )
                    return expected

            def default(expected, seq, token, tokenizer=None):
                type_, val, line, col = token
                if expected != 'EOF':
                    seq.append(val, type_, line=line, col=col)
                    return expected
                else:
                    new['wellformed'] = False
                    self._log.error(
                        'CSSUnknownRule: Expected end of rule.', token=token
                    )
                    return expected

            # unknown : ATKEYWORD S* ... ; | }
            newseq = self._tempSeq()
            wellformed, expected = self._parse(
                expected=None,
                seq=newseq,
                tokenizer=tokenizer,
                productions={
                    'CHAR': CHAR,
                    'EOF': EOF,
                    'FUNCTION': FUNCTION,
                    'INVALID': INVALID,
                    'STRING': STRING,
                    'URI': URI,
                    'S': default,  # overwrite default default!
                },
                default=default,
                new=new,
            )

            # wellformed set by parse
            wellformed = wellformed and new['wellformed']

            # post conditions
            if expected != 'EOF':
                wellformed = False
                self._log.error(
                    'CSSUnknownRule: No ending ";" or "}" found: '
                    '%r' % self._valuestr(cssText)
                )
            elif new['nesting']:
                wellformed = False
                self._log.error(
                    'CSSUnknownRule: Unclosed "{", "[" or "(": %r'
                    % self._valuestr(cssText)
                )

            # set all
            if wellformed:
                self.atkeyword = self._tokenvalue(attoken)
                self._setSeq(newseq)

    cssText = property(
        fget=_getCssText,
        fset=_setCssText,
        doc="(DOM) The parsable textual representation.",
    )

    type = property(
        lambda self: self.UNKNOWN_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    wellformed = property(lambda self: bool(self.atkeyword))



"""Testcases for cssutils.css.CSSPageRule"""

import xml.dom

import pytest

import cssutils

from . import test_cssrule


class TestMarginRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.MarginRule()
        self.rRO = cssutils.css.MarginRule(readonly=True)
        self.r_type = cssutils.css.MarginRule.MARGIN_RULE
        self.r_typeString = 'MARGIN_RULE'

    def test_init(self):
        "MarginRule.__init__()"

        r = cssutils.css.MarginRule()
        assert r.margin is None
        assert r.atkeyword is None
        assert r._keyword is None
        assert r.style.cssText == ''
        assert r.cssText == ''

        r = cssutils.css.MarginRule(margin='@TOP-left')
        assert r.margin == '@top-left'
        assert r.atkeyword == '@top-left'
        assert r._keyword == '@TOP-left'
        assert r.style.cssText == ''
        assert r.cssText == ''

        with pytest.raises(xml.dom.InvalidModificationErr):
            cssutils.css.MarginRule('@x')

    def test_InvalidModificationErr(self):
        "MarginRule.cssText InvalidModificationErr"
        # TODO: !!!

    #        self._test_InvalidModificationErr(u'@top-left')
    #        tests = {
    #            u'@x {}': xml.dom.InvalidModificationErr,
    #            }
    #        self.do_raise_r(tests)

    def test_incomplete(self):
        "MarginRule (incomplete)"
        # must be inside @page as not valid outside
        tests = {
            '@page { @top-left { ': '',  # no } and no content
            '@page { @top-left { /*1*/ ': '',  # no } and no content
            '@page { @top-left { color: red': '@page {\n    @top-left {\n        color: red\n        }\n    }',
        }
        self.do_equal_p(tests)  # parse

    def test_cssText(self):
        tests = {
            '@top-left {}': '',
            '@top-left { /**/ }': '',
            '@top-left { color: red }': '@top-left {\n    color: red\n    }',
            '@top-left{color:red;}': '@top-left {\n    color: red\n    }',
            '@top-left{color:red}': '@top-left {\n    color: red\n    }',
            '@top-left { color: red; left: 0 }': '@top-left {\n    color: red;\n    left: 0\n    }',
        }
        self.do_equal_r(tests)

        # TODO
        tests.update({
            # false selector
            #            u'@top-left { color:': xml.dom.SyntaxErr, # no }
            #            u'@top-left { color': xml.dom.SyntaxErr, # no }
            #            u'@top-left {': xml.dom.SyntaxErr, # no }
            #            u'@top-left': xml.dom.SyntaxErr, # no }
            #            u'@top-left;': xml.dom.SyntaxErr, # no }
        })

    #        self.do_raise_r(tests) # set cssText

    def test_reprANDstr(self):
        "MarginRule.__repr__(), .__str__()"
        margin = '@top-left'

        s = cssutils.css.MarginRule(margin=margin, style='left: 0')

        assert margin in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert margin == s2.margin



"""Experimental settings for special stuff."""


def set(key, value):
    """Call to enable special settings:

    ('DXImageTransform.Microsoft', True)
        enable support for parsing special MS only filter values

    Clears the tokenizer cache which holds the compiled productions!
    """
    if key == 'DXImageTransform.Microsoft' and value is True:
        from . import cssproductions, tokenize2

        tokenize2._TOKENIZER_CACHE.clear()
        cssproductions.PRODUCTIONS.insert(1, cssproductions._DXImageTransform)
